/*
 * This file contains routine for using the finder demon to exchange 
 * information.
 *
 * The interface is
 *
 * FD_Connect( )
 *     Get fd for talking to finder
 * FD_Close( )
 *     Disconnect from finder
 * FD_Install( )
 *     Install kay and name.  Adds user-specific information
 * FD_Lookup( )
 *     Lookup key and return name.  Includes timeout
 * FD_Delete( )
 *     Delete key
 */

#include "pexec.h"
#include "finder.h"
#include <stdlib.h>

/* 
.  finderhost - Name of finder host.  If null, try FINDERHOST environment
   variable.  If found, set finderport from FINDERPORT environment variable,
   if set.  If that is also null, try the default (set in 
   pexec/apps/finder/Makefile)
.  finderport - Port on finder host.  If null, try default (FINDERPORT in
   pexec/apps/finder/Makefile).  

 */
int FD_Connect( finderhost, finderport )
char *finderhost;
int  finderport;
{
    int fd;

    if (!finderport) {
	finderport = FINDERPORT;
    }
    if (!finderhost) {
	/* Try to get from environment variable */
	finderhost = getenv( "FINDERHOST" );
	if (finderhost) {
	    char *p = getenv( "FINDERPORT" );
	    if (p) finderport = atoi( p );
	}
	else {
	    finderhost = FINDERHOST;
	}
    }

    fd = PX_FD_Create_network_conn( finderhost, finderport, 1, (int *)0 );
    if (fd < 0) {
	perror( "Could not open connection" );
	return fd;
	}
    return fd;
}

int FD_Close( ctx, fd )
PX_Ctx *ctx;
int fd;
{
    close( fd );
}

#include <pwd.h>
#include <unistd.h>

static char *username = 0;

static void FD_GetUserName() 
{
    struct passwd *pent;

    if (username) return;
    pent = getpwuid( geteuid() );
    if (pent) {
	username = pent->pw_name;
    }
    else {
	username = getlogin();
	if (!username) {
	    username = "nobody";
	}
    }
}

/* 
 * Internal routines to read/write data.  These should eventually be
 * replaced by either nonblocking or explicit blocking versions.
 */
#include <errno.h>
int FD_Write( fd, buf, len )
int fd, len;
void *buf;
{
    char *p = (char *)buf;
    int nwritten = 0, n;
    while (nwritten < len) {
	n = write( fd, buf + nwritten, len - nwritten );
	if (n < 0 && (errno == EINTR || errno == EWOULDBLOCK)) continue;
	if (n < 0) {
	    perror( "FD_Write:" );
	    return -nwritten;
	}
	nwritten += n;
    }
    return nwritten;
}

/* Read EXACTLY len */
int FD_Read( fd, buf, len )
int fd, len;
void *buf;
{
    char *p = (char *)buf;
    int nread = 0, n;
    while (nread < len) {
	n = read( fd, buf + nread, len - nread );
	if (n < 0 && (errno == EINTR || errno == EWOULDBLOCK)) continue;
	if (n < 0) {
	    perror( "FD_Read:" );
	    return -nread;
	}
	nread += n;
    }
    return nread;
}

int FD_Install( ctx, fd, key, name, timeout )
PX_Ctx *ctx;
int    fd, timeout;
char   *key;
char   *name;
{
    FD_GetUserName();
    FD_Write( fd, "i ", 2 );
    FD_Write( fd, key, strlen(key) );
    FD_Write( fd, "-", 1 );
    FD_Write( fd, username, strlen( username ) );
    FD_Write( fd, " ", 1 );
    FD_Write( fd, name, strlen(name) );
    FD_Write( fd, "\n", 1 );
    
    return 0;
}

int FD_Lookup( ctx, fd, key, name, namelen )
PX_Ctx *ctx;
int    fd, namelen;
char   *key;
char   *name;
{
    int  nread, lbuflen;
    char *lbuf;

    FD_GetUserName();

    FD_Write( fd, "l ", 2 );
    FD_Write( fd, key, strlen(key) );
    FD_Write( fd, "-", 1 );
    FD_Write( fd, username, strlen( username ) );
    FD_Write( fd, "\n", 1 );
    lbuflen = strlen(key) + strlen(username) + 1 + 3;
    lbuf = (char *)malloc( lbuflen );
    /* Read a fixed amount of data */
    nread = FD_Read( fd, lbuf, lbuflen );
    /* Check that lbuf contains "key - " */
    /* Read until a newline */
    nread = 0;
    do {
	int n;
	n = read( fd, name + nread, namelen - nread );
	if (n > 0) 
	    nread += n;
    } while (nread == 0 || name[nread-1] != '\n');
    name[nread-1] = 0;
    free( lbuf );
    /* Check that name is not "not found" */
    printf( "name = %s\n", name );
    return strcmp( name, "not found" ) != 0;
}

int FD_Delete( ctx, fd, key )
PX_Ctx *ctx;
int    fd;
char   *key;
{
    FD_GetUserName();

    FD_Write( fd, "d ", 2 );
    FD_Write( fd, key, strlen(key) );
    FD_Write( fd, "-", 1 );
    FD_Write( fd, username, strlen( username ) );
    FD_Write( fd, "\n", 1 );
}

