#ifndef _PEXEC
#define _PEXEC
/*
 */

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#if defined(__STDC__) || defined(__cplusplus)
#define ANSI_ARGS(a) a
#else
#define ANSI_ARGS(a) ()
#endif

/* 
 * Forward reference of handler type
 */
typedef struct _PX_Handler *PX_Handler;


/*
 * Forward references for read and write functions
 */
typedef struct _PX_ReadFn  *PX_ReadFn;
typedef struct _PX_WriteFn *PX_WriteFn;

/* 
 * Each FD has a connection with handlers.  
 *
 * The values "read_pending" and "write_pending" should be checked when
 * attempting to read or write.  They indicate that another handler
 * has a read or write operation pending.  
 */
typedef struct {
    int            fd;         /* Unix FD for connection */
    int            in_out;     /* PX_READ for read, PX_WRITE for write */
    PX_Handler     handler;    /* Called for activity on fd */
    PX_Handler     on_eof;     /* Called on EOF */
    PX_Handler     on_err;     /* Called on error */
    int            read_pending;
    PX_ReadFn      read;       /* Called to read data */
    int            write_pending;
    PX_WriteFn     write;      /* Called to write data */
    struct timeval timeout;    /* Timeout for activity (may be infinity)
				  This is an ABSOLUTE time */
    } PX_Conn;


/*
 * Statically allocate the connections to keep the code simple.
 * It also simplifies any asynchronous modification of the data
 * structures, because no dynamic allocation is needed.
 */
typedef struct {
    int     go;                /* set to false to stop select loop */
    int     max_fd;            /* Max FD actually in use */
    PX_Conn conn[FD_SETSIZE];  /* One possible connection for each FD */
    } PX_Ctx;

/*
 * Each handler has its own definition.  Note that handlers can be 
 * chained.  This allows easy implementation of simple finite-state
 * machines, where there is a single possible (success) transition at
 * each point.  Having a prev pointer allows simple composition, with
 * any handler being able to easily remove itself.
 */
struct _PX_Handler {
    int (*fcn) ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int )); 
                               /* Function to call for handler action */
    void       *extra_data;    /* extra data (handler is passed ITSELF, 
				  however) */
    int (*info) ANSI_ARGS(( PX_Handler, char ** ));
                               /* Function to call for information on the
				  handler */
    void       *info_data;     /* extra data for info routine */
    int (*destroy) ANSI_ARGS(( PX_Handler ));
                               /* routine to free this handler */
    PX_Handler next;           /* Next handler, if any */
    PX_Handler prev;           /* Previous handler.  If null, linked
				  from a PX_Conn (if in use) */
    int        ref_count;      /* Used to keep track of use */
    int        disposition;    /* Indicates if handler should be
				  removed when an FD is closed.  Normally 
				  PX_REMOVE_ON_CLOSE */
    };

struct _PX_ReadFn {
    int (*fcn) ANSI_ARGS(( int, void *, int, PX_ReadFn ));
    void *extra_data;
    PX_ReadFn next;
    PX_ReadFn prev;
    };
struct _PX_WriteFn {
    int (*fcn) ANSI_ARGS(( int, void *, int, PX_WriteFn ));
    void *extra_data;
    PX_WriteFn next;
    PX_WriteFn prev;
    };

/*
 * Programs are also stored in a common form.
 */
typedef struct {
    int pid;
    int exit_status;
    int rd_in, wrt_out, wrt_err;   /* fd's for stdin, out, err */
    int (*on_exit)();
    void *extra_data;
    } PX_Pgm;

#define PX_FD_IS_IN_USE(a) ((a)->fd != -1)
#define PX_FD_NOT_IN_USE -1

/* This is the limit on the number of handlers for an FD */
#define PX_MAX_CHAIN_LEN 10

/* Error conditions */
#define PX_TIMEOUT 1
#define PX_EOF     2
#define PX_OTHER   100
#define PX_EXEC_FAILED 101
#define PX_CHILD_EXITED 102

/* Read/Write values */
#define PX_READ  0x1
#define PX_WRITE 0x2

/* Disposition values */
#define PX_REMOVE_ON_CLOSE 0
#define PX_KEEP_ON_CLOSE 1

/* Use these to simplify adding memory leak tracking functions */
#define MALLOC(a) malloc(a)
#define FREE(a)   free(a)

/* Use this to get the current time of day in struct timeval format.
   This isolates wacko incompatibilities.
   We generally expect tz to be a null pointer. */
#define PX_Gettimeofday(curtime,tz) gettimeofday( curtime, tz )

/* 
 * Convenience macros for invoking the read/write functions on an fd
 * as if there were read/write (but with the conn).
 */
#define PX_SYSREAD(conn,buf,n) \
    ((conn)->read->fcn)( (conn)->fd, buf, n, (conn)->read )
#define PX_SYSWRITE(conn,buf,n) \
    ((conn)->write->fcn)( (conn)->fd, buf, n, (conn)->write )

/* External interfaces, by the source file that they are defined in */
/* fds.c */
int PX_Ctx_init           ANSI_ARGS(( PX_Ctx * ));
int PX_Ctx_close_all_conn ANSI_ARGS(( PX_Ctx * ));
int PX_Ctx_end            ANSI_ARGS(( PX_Ctx * ));
PX_Handler PX_Handler_create ANSI_ARGS(( int (*)(PX_Ctx *, PX_Conn *, 
					          PX_Handler, int ), void *,
					 int (*)( PX_Handler, char **), void *,
                                         int (*)( PX_Handler) ));
PX_Handler PX_Handler_simple_create ANSI_ARGS(( int (*)(PX_Ctx *, PX_Conn *, 
					          PX_Handler, int ), void * ));
int PX_Handler_free       ANSI_ARGS(( PX_Handler ));
int PX_FD_add             ANSI_ARGS(( int, int, 
                               PX_Handler, PX_Handler, PX_Handler, 
                               struct timeval *, PX_Ctx * ));
int PX_FD_free            ANSI_ARGS(( PX_Ctx *, int ));
void PX_FD_set_active     ANSI_ARGS(( PX_Ctx *, int, int, int ));
int PX_FD_Remove_handler  ANSI_ARGS(( PX_Conn *, PX_Handler ));
/* gselect.c */
int PX_Gselect      ANSI_ARGS(( PX_Ctx *, struct timeval * ));

/* handlers.c */
int PX_Hfcn_to_stdout     ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int PX_Hfcn_to_stderr     ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int PX_FdAddToSet         ANSI_ARGS(( PX_Handler, int ));
int PX_Hfcn_fd_to_set     ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int PX_Hfcn_close_fd      ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int PX_Hfcn_error_fd      ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
int PX_Hfcn_exit_on_close ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));

/* rshstart.c */
int PX_Hfcn_handshake     ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));
void *PX_Create_handshake ANSI_ARGS(( char * ));
/*int PX_Handle_Handshake   ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int )); */
int PX_Run_on_host        ANSI_ARGS(( char *, char *, char **, 
				      int *, int *, int *, PX_Pgm ** ));

/* rdwrt.c */
int PX_Read               ANSI_ARGS(( int, void *, int, PX_ReadFn ));
int PX_Write              ANSI_ARGS(( int, void *, int, PX_WriteFn ));
int PX_Read_add           ANSI_ARGS(( PX_Ctx *, int, 
				      int (*)(int,void*,int,PX_ReadFn), 
				      void *));
int PX_Write_add          ANSI_ARGS(( PX_Ctx *, int, 
				      int (*)(int,void*,int,PX_WriteFn), 
				      void *));

/* runpgm.c */
int PX_Fork_with_pipes    ANSI_ARGS(( int *, int *, int *, int * ));
void PX_SetChildHandler   ANSI_ARGS((void));
int PX_Run_program        ANSI_ARGS(( char *, char **, 
				      int *, int *, int *, PX_Pgm ** ));

/* sockets.c */
int PX_FD_Create_network_socket ANSI_ARGS(( char *, int, int, void ** ));
int PX_FD_Create_local_socket   ANSI_ARGS(( char *, int, void ** ));
int PX_FD_Accept_connection     ANSI_ARGS(( int, void * ));
int PX_FD_Create_network_conn   ANSI_ARGS(( char *, int, int, int * ));
int PX_FD_Create_local_conn     ANSI_ARGS(( char *, int, int * ));
int PX_FD_Establish_network_conn ANSI_ARGS(( int ));
int PX_FD_Establish_local_conn  ANSI_ARGS(( int ));

/* treestart.c */

/* util.c */
struct timeval *PX_Util_Deltatime ANSI_ARGS(( int, int, struct timeval * ));
#endif

