/*
 * This code has the routines to manage FDs
 */

#include "pexec.h"
#include <memory.h>

/*@
 PX_Ctx_init - Initializes the pexec context

  Input Parameter:
. ctx - Pointer to structure to initialize

  Returns:
  0 on success.
 @*/
int PX_Ctx_init( ctx )
PX_Ctx *ctx;
{
int     i;
PX_Conn *conn;

ctx->max_fd = -1;
ctx->go     = 1;
for (i=0; i<FD_SETSIZE; i++) {
    conn		= ctx->conn + i;
    conn->fd		= PX_FD_NOT_IN_USE;
    conn->handler	= 0;
    conn->read		= 0;
    conn->read_pending	= 0;
    conn->write		= 0;
    conn->write_pending	= 0;
    conn->on_err	= 0;
    conn->on_eof	= 0;
    }
return 0;
}

/*@
  PX_Ctx_close_all_conn - Close all of the connections/fds

  Input Parameter:
. ctx - Pexec context

@*/ 
int PX_Ctx_close_all_conn( ctx )
PX_Ctx *ctx;
{
int i;
PX_Conn *conn;

for (i=0; i<=ctx->max_fd; i++) {
    conn = ctx->conn + i;
    if (PX_FD_IS_IN_USE(conn)) {
	/* This calls the handler's destroy routine */
	PX_FD_free( ctx, i );
	if (conn->fd >= 0) 
	    close( conn->fd );
	conn->fd = PX_FD_NOT_IN_USE;
	}
    }
ctx->max_fd = -1;

return 0;
}


/*@
  PX_Ctx_end - Ends all connections on a PEXEC context

  Input Parameter:
. ctx - Pointer to PEXEC structure

  Returns:
  0 on success.

.seealso: PX_Hfcn_close_on_exit
@*/
int PX_Ctx_end( ctx )
PX_Ctx *ctx;
{
return PX_Ctx_close_all_conn( ctx );
}

/*@
  PX_Ctx_go - Marks a PEXEC context for looping

  Input Parameter:
. ctx - PEXEC context

  Notes:
  This routine sets a field that is checked by 'PX_Gselect' each time
  'PX_Gselect' would call 'select'.  By using the routine 'PX_Ctx_stop',
  a routine (such as a handler) can cause 'PX_Gselect' to exit before the
  next call to 'select'.
@*/
int PX_Ctx_go( ctx )
PX_Ctx *ctx;
{
ctx->go = 1;
return 0;
}

/*@
  PX_Ctx_stop - Marks a PEXEC context for no looping

  Input Parameter:
. ctx - PEXEC context

  Notes:
  This routine sets a field that is checked by 'PX_Gselect' each time
  'PX_Gselect' would call 'select'.  By using this,
  a routine (such as a handler) can cause 'PX_Gselect' to exit before the
  next call to 'select'.

.seealso: PX_Ctx_go
@*/
int PX_Ctx_stop( ctx )
PX_Ctx *ctx;
{
ctx->go = 0;
return 0;
}

/*@
  PX_Handler_create - Creates a handler structure

  Input Parameters:
. fcn - Handler function to call.
. extra_data - pointer to extra state needed by 'fcn'.
. info - Function to call to give information on handler.  May be null.
. info_data - Pointer to extra state for 'info'
. destroy - Function to call when freeing handler.  May be null. 

  Returns:
  New handler, or null if unable to allocate space.

  Notes:
  The function 'PX_Handler_simple_create' may be used for handlers without
  special info or destroy routines.
@*/
PX_Handler PX_Handler_create( fcn, extra_data, info, info_data, destroy )
int (*fcn) ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int )); 
void *extra_data;
int (*info) ANSI_ARGS(( PX_Handler, char **));
void *info_data;
int (*destroy) ANSI_ARGS(( PX_Handler ));
{
PX_Handler new;
new = (PX_Handler)MALLOC( sizeof(struct _PX_Handler) );
if (!new) return 0;

new->fcn	 = fcn;
new->extra_data	 = extra_data;
new->info	 = info;
new->info_data	 = info_data;
new->destroy	 = destroy;
new->next	 = 0;
new->prev	 = 0;
new->ref_count	 = 1;
new->disposition = PX_REMOVE_ON_CLOSE;

return new;
}

/*@
  PX_Handler_simple_create - Creates a handler structure

  Input Parameters:
. fcn - Handler function to call.
. extra_data - pointer to extra state needed by 'fcn'.

  Returns:
  New handler, or null if unable to allocate space.

  Notes:
  This is a simple form of 'PX_Handler_create' that has the same effect 
  as
.vb
    PX_Handler_create( fcn, extra_data, 
		       (int (*)())0, (void *)0, (int (*)())0 );
.ve
@*/
PX_Handler PX_Handler_simple_create( fcn, extra_data )
int (*fcn) ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int )); 
void *extra_data;
{
return PX_Handler_create( fcn, extra_data, 
		       (int (*)())0, (void *)0, (int (*)())0 );
}

/*@
  PX_Handler_free - Frees the storage used by a handler

  Input Parameter:
. handler - Handler to free

  Notes:
  A reference counting mechanism allows a handler to be used by multiple 
  FDs.  The handler is only removed when the reference count reaches zero.

  However, extreme care must be exercised in using this feature.  In most 
  cases, it is better to create a new handler for each use, since the
  position of a handler in a chain of handlers on a single FD may differ
  from FD to FD.
@*/
int PX_Handler_free( handler )
PX_Handler handler;
{
handler->ref_count--;
if (handler->ref_count == 0) {
    if (handler->destroy)
	(*handler->destroy)( handler );
    FREE( handler );
    }
return 0;
}

/*@
 PX_FD_add - Add a file descriptor to the fds known by a pexec context

 Input Parameters:
. fd - File descriptor to add
. in_out_mask - indicates whether the file descriptor is for read ('PX_READ'),
  write ('PX_WRITE'), or both ('PXREAD | PXWRITE').
. handler - 'PX_Handler' to use for activity on 'fd'
. on_err - Handler to call for an error condition on 'fd'
. on_eof  - Handler to call for end-of-file on 'fd'
. timeout - timeout (may by a null pointer for no timeout).  Timeouts are
  in terms of absolute time.
. ctx - 'pexec' context to add 'fd' to

  Notes:
  The handlers are added at the `end` of the list of existing handlers.
@*/
int PX_FD_add( fd, in_out_mask, handler, on_err, on_eof, timeout, ctx )
int             fd, in_out_mask;
PX_Handler     handler, on_err, on_eof;
struct timeval *timeout;
PX_Ctx         *ctx;
{
PX_Conn    *conn;
PX_Handler h;
int        i;

conn		 = ctx->conn + fd;

conn->fd	 = fd;
conn->in_out     = in_out_mask;

if (conn->handler && handler) {
    h             = conn->handler;
    for (i=0; h->next && i<PX_MAX_CHAIN_LEN; i++) h = h->next;
    if (h->next) {
	return -1;
	}
    h->next       = handler;
    handler->prev = h;
    }
else {
    conn->handler	 = handler; 
    }
if (handler) handler->ref_count ++;

if (conn->on_err && on_err) {
    h            = conn->on_err;
    for (i=0; h->next && i<PX_MAX_CHAIN_LEN; i++) h = h->next;
    if (h->next) {
	return -1;
	}
    on_err->prev = h;
    }
else 
    conn->on_err	 = on_err;  
if (on_err)  on_err->ref_count ++;

if (conn->on_eof && on_eof) {
    h            = conn->on_eof;
    for (i=0; h->next && i<PX_MAX_CHAIN_LEN; i++) h = h->next;
    if (h->next) {
	return -1;
	}
    h->next      = on_eof;
    on_eof->prev = h;
    }
else
    conn->on_eof	 = on_eof;  
if (on_eof)  on_eof->ref_count ++;

if (timeout) conn->timeout	  = *timeout;
else         conn->timeout.tv_sec = 0;
if (ctx->max_fd < fd) ctx->max_fd = fd;

/* Setup the default read/write routines as required. */
if ((in_out_mask & PX_READ) && !conn->read) {
    PX_Read_add( ctx, fd, PX_Read, (void *) 0);
    }
if ((in_out_mask & PX_WRITE) && !conn->write) {
    PX_Write_add( ctx, fd, PX_Write, (void *) 0);
    }
return 0;
}

/* 
 This is an internal routine that removes the non-persistent handlers

 It returns the first handler remaining in the list
 */
PX_Handler PX_FD_remove_handlers( h )
PX_Handler h;
{
PX_Handler top, next, prev;

top  = 0;
prev = 0;
while (h) {
    next = h->next;
    if (h->disposition == PX_REMOVE_ON_CLOSE) {
	PX_Handler_free( h );
	if (prev) 
	    prev->next = next;
	if (next) 
	    next->prev = prev;
	}
    else {
	prev = h;
	if (!top) top = h;
	}
    h = next;
    }

return top;    
}

/*@
  PX_FD_free - Frees the handlers associated with an fd

  Input Parameters:
. ctx - pexec context
. fd  - File descriptor to free; this fd''s handlers will be removed.

  Notes:
  The fd is `not` closed unless the 'fd's handler''s destroy function 
  chooses to do so.  Permanent handlers are left in place.
@*/
int PX_FD_free( ctx, fd )
PX_Ctx *ctx;
int    fd;
{
PX_Conn *conn;

conn = ctx->conn + fd;

if (!PX_FD_IS_IN_USE(conn)) {
    return -1;
    }
conn->handler = PX_FD_remove_handlers( conn->handler );
conn->on_eof  = PX_FD_remove_handlers( conn->on_eof );
conn->on_err  = PX_FD_remove_handlers( conn->on_err );

conn->fd      = PX_FD_NOT_IN_USE;

return 0;
}

/*@
  PX_Hdestroy_extras - Frees any space in the handler extra data fields

  Input Parameter:
. handler - Handler to free space from

  Notes:
  This routine uses the 'free' call to free any data in the 'extra_data' 
  or 'info_data' fields.  Most handlers should call this routine (it is 
  safe even if no data is ever allocated).
@*/
int PX_Hdestroy_extras( handler )
PX_Handler handler;
{
if (handler->extra_data) {
    FREE( handler->extra_data );
    }
if (handler->info_data) {
    FREE( handler->info_data );
    }
return 0;
}

/*@
  PX_Hinfo_name - Returns the name of a handler 

  Input Parameter:
. handler - Handler to get information on 

  Output Parameter:
. name  - Pointer to a string of data

  Notes:
  The name returned by this routine was set when the routine was installed as 
  the handler
.vb
  PX_Handler_create( ..., PX_Hinfo_name, "foobar", ... );
.ve
  The string holding the name should be in permanent rather than stack storage,
  and the destroy routine 'PX_Hdestroy_extras' should `not` be used.

  This needs to be improved; this routine is a placeholder for the moment.
@*/
int PX_Hinfo_name( handler, ptr_p )
PX_Handler handler;
char       **ptr_p;
{
*ptr_p = (char *)handler->info_data;
return 0;
}  

/*@
  PX_FD_set_active - Controls whether a read/write mask is set
  
  Input Parameters:
. ctx - 'pexec' context
. fd  - fd to change active set for
. mask - Mask value to change.  Either 'PX_READ' or 'PX_WRITE'.
. flag - Value.  '1' for set, '0' for cleared.

@*/
void PX_FD_set_active( ctx, fd, mask, flag )
PX_Ctx *ctx;
int    fd, mask, flag;
{
if (flag)
    ctx->conn[fd].in_out |= mask;
else
    ctx->conn[fd].in_out &= ~mask;
}

/*@
  PX_FD_Remove_handler - Removes a single handler

  Input Parameters:
. conn  - Connection to remove handler from
. handler - handler to remove

  Notes:
  This function may be used to remove handlers that have been used to 
  
@*/
int PX_FD_Remove_handler( conn, handler )
PX_Conn    *conn;
PX_Handler handler;
{
    if (handler->prev) {
	handler->prev->next = handler->next;
	}
    else {
	conn->handler = handler->next;
	}
    if (handler->next) 
	handler->next->prev = handler->prev;
    /* Should we free the handshake structure??? */

    return PX_Handler_free( handler );
}
