/*
 * This is the ultimate select loop
 *
 * Note that it does not cache data away; that would make the code somewhat
 * faster but much more complex, as the caches would have to be updated 
 * (for example, the fd masks need to be recomputed when a new fd is added
 * or an old one removed).
 */

/* max fd is FD_SETSIZE */

/* 
 * Yet to do: convert absolute time outs to relative, or decide to 
 * only use relative.
 */

#include "pexec.h"
#include <errno.h>
/*@
  PX_Gselect - Processes activities on specified file descriptors

  Input Parameters:
. ctx - Context containing FDs to act on
. gtimeout - global timeout for this routine.  If null, there is no time
  out.  Timeouts are `absolute`, not relative.  That is, this value should
  give the time in the future when the timeout should occur, not the increment
  in the time.

  Returns:
  0 on success; else -1 with error in 'errno'.  That value corresponds to the
  system error that caused the problem.

  Return value on timeout (caused by 'gtimeout') is 0.

.seealso: PX_Ctx_init, PX_Ctx_end, PX_Ctx_stop
@*/
int PX_Gselect( ctx, gtimeout )
PX_Ctx *ctx;
struct timeval *gtimeout;
{
struct timeval timeout, elapsed, *timeoutp, curtime;
fd_set         readfds, writefds, exceptfds;
int            i, nfds;
PX_Conn        *conn;

while (ctx->go) {
    FD_ZERO( &readfds );
    FD_ZERO( &writefds );
    FD_ZERO( &exceptfds );
    /*
     * Get the active fd's 
     */
    PX_Gettimeofday( &curtime, (void *)0 );
    if (gtimeout) {
	timeout.tv_sec = gtimeout->tv_sec;
	}
    else
	timeout.tv_sec = 10000 + curtime.tv_sec;
    for (i=0; i<=ctx->max_fd; i++) {
	conn = ctx->conn + i;
	if (PX_FD_IS_IN_USE(conn)) {
	    /* Get the fd value */
	    if (conn->in_out & PX_READ) {
		FD_SET( conn->fd, &readfds );
		}
	    if (conn->in_out & PX_WRITE) {
		FD_SET( conn->fd, &writefds );
		}
	    /* Get the timeout, if any */
	    if (conn->timeout.tv_sec > 0 && 
		timeout.tv_sec > conn->timeout.tv_sec) 
		timeout.tv_sec = conn->timeout.tv_sec;
	    }
	}
    /* Use null timeout if none selected */
    timeoutp = &timeout;
    if (timeout.tv_sec == 10000 + curtime.tv_sec) {
	timeoutp = 0;
	}
    else {
	/* Convert to relative time */
	if (curtime.tv_sec > timeout.tv_sec) 
	    timeout.tv_sec = 0;
	else
	    timeout.tv_sec = timeout.tv_sec - curtime.tv_sec;
	timeout.tv_usec = 0;
	}
    /* If there are no fds, and the timeout is null, break loop */
    if (ctx->max_fd == -1 && !timeoutp) break;
    nfds = select( ctx->max_fd+1, &readfds, &writefds, &exceptfds, timeoutp );
    if (nfds == -1) {
	switch (errno) {
	    case EINTR:  /* interrupted system call */
	    break;
	    case EBADF:  /* bad fd */
	    /* This will cause all selects to fail until cleaned up */
	    return -1;

	    case EFAULT: /* bad pointer */
	    return -1;

	    case EINVAL: /* time out-of-range (tv_usec < 0 or tv_sec > 10^8 
			    or tv_usec > 10^6) */
	    return -1;

	    default: 
	    /* Unrecognized error */
	    return -1;
	    }
	continue;
	}
    if (nfds == 0) {
	/* Timeout.  Call the timeout handler for those fds */
	if (timeoutp) {
	    PX_Gettimeofday( &elapsed, (void *)0 );
	    }
	for (i=0; i<=ctx->max_fd; i++) {
	    conn = ctx->conn + i;
	    if (conn->timeout.tv_sec > 0 &&
		conn->timeout.tv_sec < elapsed.tv_sec) {
		(*conn->on_err->fcn)( ctx, conn, conn->on_err, PX_TIMEOUT );
		}
	    }
	if (gtimeout && gtimeout->tv_sec <= elapsed.tv_sec) {
	    /* global timeout for the routine.  return 0 */
	    return 0;
	    }
	}
    
    for (i=0; nfds > 0 && i<=ctx->max_fd; i++) {
	conn = ctx->conn + i;
	if (PX_FD_IS_IN_USE(conn)) {
	    if ((conn->in_out & PX_READ) && FD_ISSET( conn->fd, &readfds )) {
		nfds--;
		(*conn->handler->fcn)( ctx, conn, conn->handler, PX_READ );
		}
	    if ((conn->in_out & PX_WRITE) && FD_ISSET( conn->fd, &writefds )) {
		nfds--;
		(*conn->handler->fcn)( ctx, conn, conn->handler, PX_WRITE );
		}
	    }
	}
    }
return 0;
}
