#include "pexec.h"
#include <errno.h>
#include <memory.h>

/*
 * Some simple default handlers.  These do no processing of the data.
 * 
 * PX_ToStdout - Read data and write directly to stdout
 * PX_FdToSet  - Read and write to a collection of fds; the collection is
 *               set with PX_FdAddToSet.  Can be used to direct stdin to
 *               one or more fds.
 * PX_CloseFd  - Simply close an fd
 * PX_ErrFd    - On an error, close (same action as PX_CloseFd for now)
 * PX_ExitOnClose - Close ALL fds, invoking their on_eof handlers.
 *                  This will terminate the PX_gselect loop
 */

#define MAX_BUF 256

/*@
  PX_Hfcn_do_eof - A special handler for EOF on read

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. n - value returned by read (less than or equal to zero).

  Notes:
  This is a special handler for end-of-file and error indications from 
  a read.  A typical use from within another handler is
.vb
  n = read( ... );
  if (n <= 0)
      return PX_Hfcn_do_eof( ctx, conn, handler, n );
.ve
  This handler is normally `not` used as an file descriptors handler.
@*/
int PX_Hfcn_do_eof( ctx, conn, handler, n )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        n;
{
int err = 0;
if (n == 0) {
    /* This is eof, except for SYSV on nonblocking socket (!) */
    if (conn->on_eof)
	err = (*conn->on_eof->fcn)( ctx, conn, conn->on_eof, PX_EOF );
    }
if (n == -1) {
    if (errno == EWOULDBLOCK || errno == EAGAIN) 
	/* These indicate that no data was actually available.
	   We just ignore the read. EAGAIN is POSIX, EWOULDBLOCK is
	   BSD. */
	return 0;
    /* Other error */
    if (conn->on_err) 
	err = (*conn->on_err->fcn)( ctx, conn, conn->on_err, -1 );
    }
return err;
}

/*@
   PX_Hfcn_to_stdout - Sends data to standard output

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.
@*/
int PX_Hfcn_to_stdout( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
int n;
char buf[MAX_BUF];
    
n = (conn->read->fcn)( conn->fd, buf, MAX_BUF, conn->read );
if (n > 0) 
    write( 1, buf, n );
else 
    return PX_Hfcn_do_eof( ctx, conn, handler, n );

return 0;
}    

/*@
   PX_Hfcn_to_stderr - Sends data to standard error

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.
@*/
int PX_Hfcn_to_stderr( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
int n;
char buf[MAX_BUF];
    
n = (conn->read->fcn)( conn->fd, buf, MAX_BUF, conn->read );
if (n > 0) 
    write( 2, buf, n );
else 
    return PX_Hfcn_do_eof( ctx, conn, handler, n );

return 0;
}    

/*
 * Send data from one fd to some others.  The list of others is stored in 
 * the extra_data
 */
#define MAX_FDS 20
typedef struct _fd_list {
    int nfds;
    int fd[MAX_FDS];
    } FD_List;

/*@
  PX_FdAddToSet - Adds a file descriptor to the set of fds that a handler
  will write to

  Input Parameters:
. handler - Handler created with 'PX_Handler_create' and function 
  'PX_Hfcn_fd_to_set'
. fd - fd to write to

  Notes:
  Up to 20 fds may be specified in a set.

@*/
int PX_FdAddToSet( handler, fd )
PX_Handler handler;
int        fd;
{
FD_List *fds = (FD_List*)handler->extra_data;

if (!fds) {
    fds = (FD_List *)MALLOC( sizeof(FD_List) );
    if (!fds) return -1;
    handler->extra_data = (void *)fds;
    fds->nfds = 0;
    }
if (fds->nfds >= MAX_FDS) return -1;
fds->fd[fds->nfds++] = fd;
    
return 0;
}

/*@
   PX_Hfcn_fd_to_set - Sends data to a set of file descriptors

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.

  Notes:
  The set of file descriptors to which data is sent is set with the
  routine 'PX_FdAddToSet'.

@*/
int PX_Hfcn_fd_to_set( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
FD_List *fds = (FD_List*)handler->extra_data;
int     n, i, maxbuf = MAX_BUF;
char buf[MAX_BUF];

n = (conn->read->fcn)( conn->fd, buf, maxbuf, conn->read );
if (n > 0) {
    for (i=0; i<fds->nfds; i++) {
	write( fds->fd[i], buf, n );
	}
    }
else 
    return PX_Hfcn_do_eof( ctx, conn, handler, n );

return 0;
}

/*@
   PX_Hfcn_close_fd - Closes an fd

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.
@*/
int PX_Hfcn_close_fd( ctx, conn, handler, err )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        err;
{
close( conn->fd );
conn->fd = -1;

return 0;
}

/*@
   PX_Hfcn_error_fd - Closes an fd on error

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.
@*/
int PX_Hfcn_error_fd( ctx, conn, handler, err )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        err;
{
/* Attempt to close */
close( conn->fd );
conn->fd = -1;

return 0;
}

/*@
   PX_Hfcn_exit_on_close - Closes all fds and exit PX_gselec

  Input Parameters:
. ctx - Pexec context
. conn - Connection
. handler - This handler 
. in_out - In/out mask.  Ignored.

.seealso: PX_Ctx_end
@*/
int PX_Hfcn_exit_on_close( ctx, conn, handler, err )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        err;
{
close( conn->fd );
conn->fd = -1;

return PX_Ctx_close_all_conn( ctx );
}
