/*
 * These routines handle setting up a master and connecting to it
 */
#include "pexec.h"
#include <errno.h>
#include "msgpass.h"
/* For htonl/ntohl etc. */
#include <sys/types.h>
#include <netinet/in.h>
#include <stdio.h>

/* Local definitions only for setting up connections */
typedef struct {
    void *sock_info;
    MSG_State *mctx;
    } Connection_info;

static int dbg_flag = 1;

/* 
 * Debugging macros
 */
#define DBG_MSG(mctx,str) {if ((mctx)->dbg_flags) printf( "%s\n", str );}

/* Forward references */
void MSG_Group_init ANSI_ARGS(( MSG_State *, int ));
int MSG_Hfcn_setgroup ANSI_ARGS(( PX_Ctx *, PX_Conn *, PX_Handler, int ));

/* Group 0 message types, used by initialization routines */
#define MSG_0_PING    1
#define MSG_0_EXIT    2
#define MSG_0_GRPINFO 3
#define MSG_0_STDIN   4
#define MSG_0_STDOUT  5
#define MSG_0_STDERR  6
#define MSG_0_SIGNAL 10

/* 
   This is called when a listen allows an accept 
   
   This function installs a handshake handler and the message-passing
   receive handler on the new fd, along with a handshake timelimit.
 
 */
int MSG_GetConnection( ctx, conn, handler, in_mask )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_mask;
{
int fd;
PX_Handler err_fd, close_fd, fd_handshake, fd_msgpass, fd_setgroup;
MSG_FD *fdinfo;
Connection_info *cinfo;
    
cinfo = (Connection_info *)handler->extra_data;
DBG_MSG(cinfo->mctx,"Accepting a connection")
fd = PX_FD_Accept_connection( conn->fd, cinfo->sock_info );
if (fd < 0) {
    fprintf( stderr, "accept errno = %d\n", errno );
    perror( "Accept failed" );
    return fd;
    }
DBG_MSG(cinfo->mctx,"Accept succeeded")

fdinfo = (MSG_FD *)MALLOC( sizeof(MSG_FD) );
if (!fdinfo) {
    return -1;
    }
memset( fdinfo, 0, sizeof(MSG_FD) );
fdinfo->fdin.fd	   = fd;
fdinfo->fdin.mctx  = cinfo->mctx;
fdinfo->fdout.fd   = fd;
fdinfo->fdout.mctx = cinfo->mctx;
fdinfo->am_master  = 1;
fdinfo->state      = 1;

close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
fd_handshake  = PX_Handler_simple_create( PX_Hfcn_handshake, 
				   PX_Create_handshake("MsgPass") );
PX_FD_add( fd, PX_READ | PX_WRITE, fd_handshake, err_fd, close_fd, 
	  PX_Util_Deltatime( 10, 0, (void *)0 ), ctx );
close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
fd_setgroup   = PX_Handler_simple_create( MSG_Hfcn_setgroup, (void *)fdinfo );
PX_FD_add( fd, PX_READ | PX_WRITE, fd_setgroup, err_fd, close_fd, 
	  PX_Util_Deltatime( 30, 0, (void *)0 ), ctx );
close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
fd_msgpass    = PX_Handler_simple_create( MSG_Hfcn_incoming, (void *)fdinfo );
PX_FD_add( fd, PX_READ | PX_WRITE, fd_msgpass, err_fd, close_fd, 
	  (struct timeval *)0, ctx );
PX_FD_set_active( ctx, fd, PX_WRITE, 0 );

return 0;
}

int MSG_Master_setup( ctx, portnum, mctx )
PX_Ctx    *ctx;
int       *portnum;
MSG_State **mctx;
{
void       *sock_info;
int        fd;
PX_Handler getconnection;
MSG_Group  group;
static     Connection_info cinfo;

/* Create an initial mctx and return it */
*mctx = (MSG_State *) MALLOC( sizeof(MSG_State) );
if (!*mctx) {
    return -1;
    }
memset( (*mctx), 0, sizeof(MSG_State) );
(*mctx)->dbg_flags = dbg_flag;
MSG_Group_init( (*mctx), 0 );
group = (*mctx)->groups[0];
/* Setup the master group */
group->n_members = 1;
group->lrank	 = 0;

cinfo.mctx = *mctx;

if (*portnum <= 0) {
    *portnum = 1235;
    }

/* Add to ctx a network listener for connections */
DBG_MSG(*mctx,"Creating a network socket")
fd = PX_FD_Create_network_socket( (char *)0, *portnum, 1, &cinfo.sock_info );
if (fd < 0) {
    fprintf( stderr, "Error creating network socket\n" );
    }
else {
    DBG_MSG(*mctx,"Adding a connection handler")
    getconnection = PX_Handler_simple_create( MSG_GetConnection, &cinfo );
    PX_FD_add( fd, PX_READ, getconnection, (void *)0, (void *)0, 
	      (struct timeval *)0, ctx );
    PX_Handler_free( getconnection );
    }

/* Save information on contacting the master (us) */
(*mctx)->mastername = 0;
(*mctx)->masterport = *portnum;

return 0;
}

/*
 * If mctx is not set, create it.
 */
int MSG_Master_connect( ctx, host, port, mctx )
PX_Ctx *ctx;
char   *host;
int    port;
MSG_State **mctx;
{
int        fd;
PX_Conn    *conn;
PX_Handler err_fd, close_fd, fd_msgpass, fd_setgroup;
MSG_FD     *fdinfo;

/* Open a socket connection for the connection to host/port; return
   an initial mctx */

/* Create the MSG context */
if (!*mctx) {
    *mctx = (MSG_State *) MALLOC( sizeof(MSG_State) );
    if (!*mctx) {
	return -1;
	}
    memset( (*mctx), 0, sizeof(MSG_State) );
    (*mctx)->dbg_flags = dbg_flag;
    }
/* Create the fdinfo context */
fdinfo = (MSG_FD *)MALLOC( sizeof(MSG_FD) );
if (!fdinfo) {
    return -1;
    }

DBG_MSG(*mctx,"Connecting to remote host")
fd = PX_FD_Create_network_conn( host, port, 1, (int *)0 );
if (fd < 0) {
    perror( "Could not open connection" );
    return 1;
    }
DBG_MSG(*mctx,"Got connection")

memset( fdinfo, 0, sizeof(MSG_FD) );
fdinfo->fdin.fd	   = fd;
fdinfo->fdin.mctx  = *mctx;
fdinfo->fdout.fd   = fd;
fdinfo->fdout.mctx = *mctx;
fdinfo->am_master  = 0;
fdinfo->state      = 1;

/* Add fd to ctx */
close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
fd_setgroup   = PX_Handler_simple_create( MSG_Hfcn_setgroup, (void *)fdinfo );
PX_FD_add( fd, PX_READ | PX_WRITE, fd_setgroup, err_fd, close_fd, 
	  PX_Util_Deltatime( 30, 0, (void *)0 ), ctx );

close_fd      = PX_Handler_simple_create( PX_Hfcn_close_fd, (void *)0 );
err_fd        = PX_Handler_simple_create( PX_Hfcn_error_fd, (void *)0 );
fd_msgpass    = PX_Handler_simple_create( MSG_Hfcn_incoming, (void *)fdinfo );
PX_FD_add( fd, PX_READ | PX_WRITE, fd_msgpass, err_fd, close_fd, 
	  (struct timeval *)0, ctx );
PX_FD_set_active( ctx, fd, PX_WRITE, 0 );

/* Save information on contacting the master */
(*mctx)->mastername = (char *)MALLOC( strlen(host) + 1 );
strcpy( (*mctx)->mastername, host );
(*mctx)->masterport = port;

/* Now need to start a handshake on the connection */
DBG_MSG(*mctx,"Doing handshake with remote host")
conn = &ctx->conn[fd];
if (PX_SYSWRITE( conn, "MsgPass", 7 ) != 7) {
    /* Failed to send handshake; close and exit */
    printf( "Need to clean up FD!\n" );
    close( fd );
    return 0;
    }
if (PX_SYSWRITE( conn, "Start", 5 ) != 5) {
    /* Failed to send handshake; close and exit */
    printf( "Need to clean up FD!\n" );
    close( fd );
    return 0;
    }

/* We still need to setup the basic group */
MSG_Group_init( *mctx, 0 );
return 0;
}

/*
 * Here are the MSG group management routines.  
 *
 * To start with, we create group 0
 */

void MSG_Group_init( mctx, group_id )
MSG_State *mctx;
int       group_id;
{
MSG_Group group;
int       i;

/* Special case for group_id == 0 */
group = mctx->groups[group_id];
if (!group) {
    group = (MSG_Group) MALLOC( sizeof(struct _MSG_Group) );
    mctx->groups[group_id] = group;
    }

/* Common code */
for (i=0; i<MSG_MAX_TAG; i++) {
    group->fcn[i] = 0;
    group->extra_data[i] = 0;
    }
for (i=0; i<MSG_MAX_PROCS; i++) {
    group->l_to_g[i] = -1;
    }
if (group_id == 0) {
    /* These need to be set with the correct functions to call */
    group->fcn[MSG_0_PING]    = 0;
    group->fcn[MSG_0_EXIT]    = 0;
    group->fcn[MSG_0_GRPINFO] = 0;
    group->fcn[MSG_0_STDIN]   = 0;
    group->fcn[MSG_0_STDOUT]  = 0;
    group->fcn[MSG_0_STDERR]  = 0;
    group->fcn[MSG_0_SIGNAL]  = 0;
    }
}

/*
 * These routines help display the state of the message passing system
 */
void MSG_Dbg_state( mctx )
MSG_State *mctx;
{
int       i, j;
MSG_FD    *fdinfo;
MSG_Group group;
printf( "Msg state:\ndbg_flags\t%d\nn_conn  \t%d\n", mctx->dbg_flags, 
        mctx->n_conn );
for (i=0; i<mctx->n_conn; i++) {
    fdinfo = mctx->fdinfo[i];
    printf( " Connection %d\n", i );
    if (fdinfo) {
	printf( " FDin:\n  busy\t%d\n  fd  \t%d\n  n_read\t%d\n", 
	       fdinfo->fdin.busy, 
	       fdinfo->fdin.fd,
	       fdinfo->fdin.n_read );
	printf( " FDout:\n  busy\t%d  fd  \t%d\n  n_written\t%d\n", 
	       fdinfo->fdout.busy, 
	       fdinfo->fdout.fd,
	       fdinfo->fdout.n_written );
	}
    else {
	printf( "No fdinfo!\n" );
	}
    }
for (i=0; i<MSG_MAX_GROUPS; i++) {
    group = mctx->groups[i];
    if (group) {
	printf( " Group %d\n size \t%d\n lrank \t%d\n group_id\t%d\n",
	        i, group->n_members, group->lrank, group->group_id );
	for (j=0; j<group->n_members; j++) {
	    printf( "%d ", group->l_to_g[j] );
	    }
	printf( "\n" );
	}
    }
if (mctx->mastername) {
    printf( " Master is %s at port %d\n", mctx->mastername, mctx->masterport );
    }
else {
    printf( " No mastername/port!\n" );
    }
}

/*
 * This is the handler that is called to complete the handshake that forms
 * the groups.  This function serves as both the master and slave just
 * to keep the code together.
 *
 * Master                                      Slave
 *                                             Sends ("Start")
 * Recvs("Start")
 * Sends Group0 size, slaves'rank
 *                                             Recvs size, rank
 */
int MSG_Hfcn_setgroup( ctx, conn, handler, in_out )
PX_Ctx     *ctx;
PX_Conn    *conn;
PX_Handler handler;
int        in_out;
{
MSG_FD     *fdinfo = (MSG_FD *)handler->extra_data;
MSG_State  *mctx = fdinfo->fdin.mctx;
MSG_Group  group = mctx->groups[0];
char       buf[20];
int        n, lrank;
struct     { int size:32, lrank:32; } ibuf;

if (fdinfo->am_master) {
    switch (fdinfo->state) {
	case 1: 
	    n = PX_SYSREAD( conn, buf, 5 );
	    if (n != 5) {
		close( conn->fd );
		}
	    DBG_MSG( mctx, "Got 'send group info' msg" )
	    fdinfo->state = 2;
	    lrank      = group->n_members++;
	    ibuf.lrank = htonl( lrank );
	    ibuf.size  = htonl( group->n_members );
	    n = PX_SYSWRITE( conn, &ibuf, sizeof(ibuf) );
	    if (n != sizeof(ibuf)) {
		close( conn->fd );
		}
	    /* Update the tables */
	    mctx->fdinfo[lrank] = fdinfo;
	    group->l_to_g[lrank] = lrank;
	    if (lrank >= mctx->n_conn) mctx->n_conn = lrank + 1;
	    
	    /* Inform others of the new member - How to do this 
	       in a reliable way? */

	    /* Done with setup of group */
	    PX_FD_Remove_handler( conn, handler );
	    break;

	/* Case 2 for additional data later */
        default:
	    close( conn->fd );
	}
    }
else {
    switch (fdinfo->state) {
	case 1:
	    n = PX_SYSREAD( conn, &ibuf, sizeof(ibuf) );
	    DBG_MSG( mctx, "Read group information" )
	    group->n_members = ntohl( ibuf.size );
	    group->lrank     = lrank = ntohl( ibuf.lrank );
	    fdinfo->state = 2;
	    /* Update the tables */
	    mctx->fdinfo[0] = fdinfo;
	    group->l_to_g[0] = 0;
	    if (lrank >= mctx->n_conn) mctx->n_conn = lrank + 1;

	    /* Done with setup of groups */
	    PX_FD_Remove_handler( conn, handler );
	    break;
        default:	    
	    close( conn->fd );
	}
    }
}
