#include "pexec.h"
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>

/* 
 * We have a simple, static list of active pids.  A value of -1
 * indicates an available slot; to shorten searches, "npgms" is
 * a highwater mark.  
 * The signal handler for SIGCHLD here is the counterpart to the PX_gselect
 * ALL child signals are handled here.
 */
#define MAX_PGMS 128
static int     npgms = 0;
static PX_Pgm  pgms[MAX_PGMS];

/* This is used to connect with the PX_Gselect context */
static PX_Ctx  *savepx = 0;

/* Do a fork, and return pid, and for parent,  fds for stdin, stdout, and
   stderr */
/*@
  PX_Proc_fork_with_pipes - Fork a process and setup pipes for stdin/out/err

  Output Parameters:
. pid   - pid of created process.
. rd_in - fd for standard input to program
. wrt_out - fd for standard output from program
. wrt_err - fd for standard error from program

@*/
int PX_Proc_fork_with_pipes( pid, rd_in, wrt_out, wrt_err )
int *pid, *rd_in, *wrt_out, *wrt_err;
{
int read_out_pipe[2], write_in_pipe[2], read_err_pipe[2];

if((pipe(read_out_pipe) == -1) || (pipe(write_in_pipe) == -1) ||
   (pipe(read_err_pipe) == -1))
    return -1;

/* Create a new job */
*pid = fork();
switch (*pid) {
    case 0:  /* Child */
    /* Close the unused portions of the pipes, and the corresponding
       stdin/out/err fds */
    close(read_out_pipe[0]);
    close(write_in_pipe[1]);
    close(read_err_pipe[0]);

    /* STDERR */
    close(2);               
    dup2(read_err_pipe[1],2);
    close(read_err_pipe[1]); 

    /* STDOUT */
    close(1);               
    dup2(read_out_pipe[1],1);
    close(read_out_pipe[1]); 

    /* STDIN */
    close(0);                
    dup2(write_in_pipe[0],0);
    close(write_in_pipe[0]); 

    return 0;
	   
    case -1:
    return -1;
    
    default:
    /* Close unused portion of pipes */
    close(read_out_pipe[1]); 
    close(read_err_pipe[1]);
    close(write_in_pipe[0]);
 
    *wrt_out = read_out_pipe[0];
    *wrt_err = read_err_pipe[0];
    *rd_in   = write_in_pipe[1];
    } /* End of switch on select for fork return */
return 0;
}


/* Set SIGCHLD handler */

RETSIGTYPE PX_Proc_do_sigchild( sig, code, scp )
int               sig, code;
struct sigcontext *scp;
/* Some systems have an additional char *addr */
/* POSIX may have only sig ! */
{
int prog_stat, pid;
int i, j;

/* Really need to block further signals until done ... */
/* fprintf( stderr, "Got SIGCHLD...\n" ); */
while (1) {
    pid	   = waitpid( (pid_t)(-1), &prog_stat, WNOHANG );
    if (pid == -1 && errno == EINTR) continue;

    if (pid <= 0) break;

    if (npgms > 0 && pid && 
	(WIFEXITED(prog_stat) || WIFSIGNALED(prog_stat))) {
	/* The child has stopped. Remove it from the jobs array */
	for (i = 0; i<npgms; i++) {
	    if (pgms[i].pid > 0 && pgms[i].pid == pid) {
		pgms[i].pid = 0;
		if (WIFEXITED(prog_stat)) {
		    if (pgms[i].on_exit) {
			(*pgms[i].on_exit)( pid, WEXITSTATUS(prog_stat),
					   pgms[i].extra_data );
			}
		    break;
		    }
		}
	    }
	}
    }
}

/* 
   This routine sets the signal handler for SIGCHLD events so that

   the signal handler is NOT reset
   SIGCHLD is NOT delivered while processing SIGCHLD

   Note that using a simple "signal" is unreliable because some signal
   implementations reset the signal handler to the default on delivery (!)
 */
/*@
  PX_Proc_set_child_handler - Setup a signal handler for SIGCHLD events

  Input Parameter:
. ctx - Pexec context

  Notes:
  The signal handler installed by this routine will attempt to call the
  appropriate handlers for closing the standard in/out/err pipes to the
  process, if these fds are known by the pexec context ('ctx').  The
  process exit handler will also be called (see the 'PX_pgm' structure).
@*/
void PX_Proc_set_child_handler( ctx )
PX_Ctx *ctx;
{
struct sigaction oldact;

/* Save the pexec context */
savepx = ctx;

sigaction( SIGCHLD, (struct sigaction *)0, &oldact );
oldact.sa_handler = PX_Proc_do_sigchild;

/* AIX 3.2.x does not have SA_RESETHAND */
#ifdef SA_RESETHAND
oldact.sa_flags   = oldact.sa_flags & ~(SA_RESETHAND);
#endif

sigaddset( &oldact.sa_mask, SIGCHLD );
sigaction( SIGCHLD, &oldact, (struct sigaction *)0 );
}

/*
 * This is a simple exit handler to close the pipes opened by 
 * PX_Fork_with_pipes.
 */
int PX_Proc_close_run_fds( pid, exit_stat, extra_data )
int pid, exit_stat;
void *extra_data;
{
PX_Pgm *pgm = (PX_Pgm *)extra_data;
PX_Conn *conn;

/* This invokes the PX err handlers for the fds */

if (pgm->rd_in >= 0) {
    conn = savepx->conn + pgm->rd_in;
    if (PX_FD_IS_IN_USE(conn)) {
	if (conn->on_err) {
	    (*conn->on_err->fcn)( savepx, conn, conn->on_err, 
				  PX_CHILD_EXITED );
	    }
	else {
	    close( pgm->rd_in );
	    }
	conn->fd = PX_FD_NOT_IN_USE;
	}
    else
	close( pgm->rd_in );
    }

/*
 * By default, we don't do anything with the other pipes, since there may 
 * still be data in them.  
 */
#ifdef FOO
if (pgm->wrt_out >= 0) {
    conn = savepx->conn + pgm->wrt_out;
    if (PX_FD_IS_IN_USE(conn)) {
	/* */
	if (conn->on_err) {
	    (*conn->on_err->fcn)( savepx, conn, conn->on_err, 
				  PX_CHILD_EXITED );
	    }
	else {
	    close( pgm->wrt_out );
	    }
	conn->fd = PX_FD_NOT_IN_USE;
	}
    else
	close( pgm->wrt_out );
    }

if (pgm->wrt_err >= 0) {
    conn = savepx->conn + pgm->wrt_err;
    if (PX_FD_IS_IN_USE(conn)) {
	if (conn->on_err) {
	    (*conn->on_err->fcn)( savepx, conn, conn->on_err, 
				  PX_CHILD_EXITED );
	    }
	else {
	    close( pgm->wrt_err );
	    }
	conn->fd = PX_FD_NOT_IN_USE;
	}
    else
	close( pgm->wrt_err );
    }
#endif

return 0;
}

/*
 * To run a program:
 *
 * Fork with pipes, install pid in the table (with an exit handler)
 * If parent,
 *    Add the pipes to the select context, with the appropriate handlers
 * If child, 
 *    just execvp the program
 * The exit handler closes the pipes as necessary.
 *
 * This version RETURNS the fds for stdin/out/err for the program.
 * As a subsequent step, install the appropriate handlers.
 *
 * Returns PX_Pgm pointer
 */

/*@
  PX_Proc_run_program - Runs a program on this processor

  Input Parameters:
. pgm  - Name of program
. argv - Argument vector; null terminated

  Output Parameters:
. rd_in - fd for standard input to program
. wrt_out - fd for standard output from program
. wrt_err - fd for standard error from program
. pgmst  - Process description structure

 Notes:
 This process returns as soon as the fork to create the process to run the
 program in completes.
@*/
int PX_Proc_run_program( pgm, argv, rd_in, wrt_out, wrt_err, pgmst )
char *pgm;
char **argv;
int *rd_in, *wrt_out, *wrt_err;
PX_Pgm **pgmst;
{
int err, pid;

*pgmst = 0;
err = PX_Proc_fork_with_pipes( &pid, rd_in, wrt_out, wrt_err );
if (err) return err;
if (pid) {
    /* Parent */
    pgms[npgms].on_exit	   = PX_Proc_close_run_fds;
    pgms[npgms].rd_in	   = *rd_in;
    pgms[npgms].wrt_out	   = *wrt_out;
    pgms[npgms].wrt_err	   = *wrt_err;
    pgms[npgms].extra_data = (void *)&pgms[npgms];
    /* Set pid LAST just in case we get a signal about this child */
    pgms[npgms].pid	   = pid;
    *pgmst                 = pgms + npgms;
    npgms++;
    }
else {
    /* Child */
    err  = execvp( pgm, argv );
    /* execvp does not return on success */
    err  = PX_EXEC_FAILED;
    }
return err;
}
